# -*- coding:utf-8 -*-
"""
@Author  :   g1879
@Contact :   g1879@qq.com
@File    :   downloadKit.py
"""
from pathlib import Path
from queue import Queue
from re import sub
from threading import Thread, Lock
from time import sleep, perf_counter
from typing import Union, Tuple
from urllib.parse import quote, urlparse

from DataRecorder import Recorder
from requests import Session, Response
from requests.structures import CaseInsensitiveDict

from ._funcs import FileExistsSetter, PathSetter, BlockSizeSetter, copy_session, _set_charset, _get_file_info, LogMode
from .mission import Task, Mission, MissionData, BaseTask


class DownloadKit(object):
    file_exists = FileExistsSetter()
    goal_path = PathSetter()
    block_size = BlockSizeSetter()

    def __init__(self,
                 goal_path: Union[str, Path] = None,
                 roads: int = 10,
                 session=None,
                 file_exists: str = 'rename'):
        """初始化                                                                         \n
        :param goal_path: 文件保存路径
        :param roads: 可同时运行的线程数
        :param session: 使用的Session对象，或配置对象、页面对象等
        :param file_exists: 有同名文件名时的处理方式，可选 'skip', 'overwrite', 'rename'
        """
        self._roads = roads
        self._missions = {}
        self._threads = {i: None for i in range(self._roads)}
        self._waiting_list: Queue = Queue()
        self._missions_num = 0
        self._stop_printing = False  # 用于控制显示线程停止
        self._lock = Lock()
        self._page = None  # 如果接收页面对象则存放于此
        self._retry = None
        self._interval = None
        self._timeout = None

        self._print_mode = None
        self._log_mode = None
        self._logger = None

        self.goal_path: str = goal_path or '.'
        self.file_exists: str = file_exists
        self.show_errmsg: bool = False
        self.block_size: Union[str, int] = '50M'  # 分块大小
        self.session = session

    def __call__(self,
                 file_url: str,
                 goal_path: Union[str, Path] = None,
                 rename: str = None,
                 file_exists: str = None,
                 post_data: Union[str, dict] = None,
                 show_msg: bool = True,
                 **kwargs) -> tuple:
        """以阻塞的方式下载一个文件并返回结果，主要用于兼容旧版DrissionPage                                     \n
        :param file_url: 文件网址
        :param goal_path: 保存路径
        :param rename: 重命名的文件名
        :param file_exists: 遇到同名文件时的处理方式，可选 'skip', 'overwrite', 'rename'，默认跟随实例属性
        :param post_data: post方式使用的数据
        :param show_msg: 是否打印进度
        :param kwargs: 连接参数
        :return: 任务结果和信息组成的tuple
        """
        return self.add(file_url=file_url,
                        goal_path=goal_path,
                        rename=rename,
                        file_exists=file_exists,
                        post_data=post_data,
                        split=False,
                        **kwargs).wait(show=show_msg)

    def set_print(self) -> LogMode:
        """设置打印到控制台的信息，可选全部、错误任务、不打印"""
        if self._print_mode is None:
            self._print_mode = LogMode()
        return self._print_mode

    def set_log(self, log_path: [str, Path] = None) -> LogMode:
        """设置记录到文件的信息，可选全部、错误任务、不记录   \n
        :param log_path: 记录文件路径
        :return: LogMode对象
        """
        if self._log_mode is None:
            self._log_mode = LogMode()
        if log_path is not None:
            if self._logger is not None:
                self._logger.set_path(log_path)
            else:
                self._logger = Recorder(log_path, 1)
        return self._log_mode

    @property
    def roads(self) -> int:
        """可同时运行的线程数"""
        return self._roads

    @roads.setter
    def roads(self, val: int) -> None:
        """设置可同时运行的线程数"""
        if self.is_running:
            print('有任务未完成时不能改变roads。')
            return
        if val != self._roads:
            self._roads = val
            self._threads = {i: None for i in range(self._roads)}

    @property
    def retry(self) -> int:
        """返回连接失败时重试次数"""
        if self._retry is not None:
            return self._retry
        elif self._page is not None:
            return self._page.retry_times
        else:
            return 3

    @retry.setter
    def retry(self, times: int) -> None:
        """设置连接失败时重试次数"""
        if not isinstance(times, int) or times < 0:
            raise TypeError('times参数只能接受int格式且不能小于0。')
        self._retry = times

    @property
    def interval(self) -> float:
        """返回连接失败时重试间隔"""
        if self._interval is not None:
            return self._interval
        elif self._page is not None:
            return self._page.retry_interval
        else:
            return 5

    @interval.setter
    def interval(self, seconds: Union[int, float]) -> None:
        """设置连接失败时重试间隔"""
        if not isinstance(seconds, (int, float)) or seconds < 0:
            raise TypeError('seconds参数只能接受int或float格式且不能小于0。')
        self._interval = seconds

    @property
    def timeout(self) -> float:
        """返回连接超时时间"""
        if self._timeout is not None:
            return self._timeout
        elif self._page is not None:
            return self._page.timeout
        else:
            return 20

    @timeout.setter
    def timeout(self, seconds: Union[int, float]) -> None:
        """设置连接超时时间"""
        if not isinstance(seconds, (int, float)) or seconds < 0:
            raise TypeError('seconds参数只能接受int或float格式且不能小于0。')
        self._timeout = seconds

    @property
    def waiting_list(self) -> Queue:
        """返回等待队列"""
        return self._waiting_list

    @property
    def session(self) -> Session:
        return self._session

    @session.setter
    def session(self, session) -> None:
        try:
            from DrissionPage import WebPage, MixPage, Drission, SessionPage, SessionOptions

            if isinstance(session, SessionOptions):
                self._session = Drission(driver_or_options=False, session_or_options=session).session
            elif isinstance(session, Drission):
                self._session = session.session
            elif isinstance(session, (SessionPage, WebPage, MixPage)):
                self._session = session.session
                self._page = session
            else:
                self._session = Drission(driver_or_options=False).session

        except ImportError:
            self._session = Session()

    @property
    def is_running(self) -> bool:
        """返回是否有线程还在运行"""
        return any(self._threads.values()) or not self.waiting_list.empty()

    def set_proxies(self, http: str = None, https: str = None) -> None:
        """设置代理地址及端口，例：'http://127.0.0.1:1080'         \n
        :param http: http代理地址及端口
        :param https: https代理地址及端口
        :return: None
        """
        if not http.startswith('http://'):
            http = f'http://{http}'
        if not https.startswith('http'):
            https = f'http://{https}'
        self._session.proxies = {'http': http, 'https': https}

    def add(self,
            file_url: str,
            goal_path: Union[str, Path] = None,
            rename: str = None,
            file_exists: str = None,
            post_data: Union[str, dict] = None,
            split: bool = True,
            **kwargs) -> Mission:
        """添加一个下载任务并将其返回                                                                    \n
        :param file_url: 文件网址
        :param goal_path: 保存路径
        :param rename: 重命名的文件名
        :param file_exists: 遇到同名文件时的处理方式，可选 'skip', 'overwrite', 'rename'，默认跟随实例属性
        :param post_data: post方式使用的数据
        :param split: 是否允许多线程分块下载
        :param kwargs: 连接参数
        :return: 任务对象
        """
        data = MissionData(url=file_url,
                           goal_path=str(goal_path or self.goal_path),
                           rename=rename,
                           file_exists=file_exists or self.file_exists,
                           post_data=post_data,
                           split=split,
                           kwargs=kwargs)
        self._missions_num += 1
        mission = Mission(self._missions_num, data, self)
        self._missions[self._missions_num] = mission
        self._run_or_wait(mission)
        return mission

    def _run_or_wait(self, mission: BaseTask):
        """接收任务，有空线程则运行，没有则进入等待队列"""
        thread_id = self._get_usable_thread()
        if thread_id is not None:
            thread = Thread(target=self._run, args=(thread_id, mission))
            self._threads[thread_id] = {'thread': thread, 'mission': None}
            thread.start()
        else:
            self._waiting_list.put(mission)

    def _run(self, ID: int, mission: BaseTask) -> None:
        """线程函数                                 \n
        :param ID: 线程id
        :param mission: 任务对象，Mission或Task
        :return: None
        """
        while True:
            if not mission:  # 如果没有任务，就从等候列表中取一个
                if not self._waiting_list.empty():
                    try:
                        mission = self._waiting_list.get(True, .5)
                    except Exception:
                        self._waiting_list.task_done()
                        break
                else:
                    break

            self._threads[ID]['mission'] = mission
            self._download(mission, ID)
            mission = None

        self._threads[ID] = None

    def get_mission(self, mission_or_id: Union[int, Mission]) -> Mission:
        """根据id值获取一个任务                 \n
        :param mission_or_id: 任务或任务id
        :return: 任务对象
        """
        return self._missions[mission_or_id] if isinstance(mission_or_id, int) else mission_or_id

    def get_failed_missions(self) -> list:
        """返回失败任务列表"""
        return [i for i in self._missions.values() if i.result is False]

    def wait(self,
             mission: Union[int, Mission] = None,
             show: bool = False,
             timeout: float = None) -> Union[tuple, None]:
        """等待所有或指定任务完成                                    \n
        :param mission: 任务对象或任务id，为None时等待所有任务结束
        :param show: 是否显示进度
        :param timeout: 超时时间，默认为连接超时时间，0为无限
        :return: 任务结果和信息组成的tuple
        """
        timeout = timeout if timeout is not None else self.timeout
        if mission:
            return self.get_mission(mission).wait(show, timeout)

        else:
            if show:
                self.show(False)
            else:
                t1 = perf_counter()
                while self.is_running or (perf_counter() - t1 < timeout or timeout == 0):
                    sleep(0.1)

    def cancel(self):
        """取消所有等待中或执行中的任务"""
        for m in self._missions.values():
            m.cancel()

    def show(self, asyn: bool = True, keep: bool = False) -> None:
        """实时显示所有线程进度                 \n
        :param asyn: 是否以异步方式显示
        :param keep: 任务列表为空时是否保持显示
        :return: None
        """
        if asyn:
            Thread(target=self._show, args=(2, keep)).start()
        else:
            self._show(0.1, keep)

    def _show(self, wait: float, keep: bool = False) -> None:
        """实时显示所有线程进度"""
        self._stop_printing = False

        if keep:
            Thread(target=self._stop_show).start()

        t1 = perf_counter()
        while not self._stop_printing and (keep or self.is_running or perf_counter() - t1 < wait):
            print(f'\033[K', end='')
            print(f'等待任务数：{self._waiting_list.qsize()}')
            for k, v in self._threads.items():
                m = v['mission'] if v else None
                if m:
                    items = (m.mission.rate, m.mid) if isinstance(m, Task) else (m.rate, m.id)
                    path = f'M{items[1]} {items[0]}% {m}'
                else:
                    path = '空闲'
                print(f'\033[K', end='')
                print(f'线程{k}：{path}')

            print(f'\033[{self.roads + 1}A\r', end='')
            sleep(0.4)

        print(f'\033[1B', end='')
        for i in range(self.roads):
            print(f'\033[K', end='')
            print(f'线程{i}：空闲')

        print()

    def _connect(self,
                 url: str,
                 mode: str = 'get',
                 data: Union[dict, str] = None,
                 **kwargs) -> Tuple[Union[Response, None], str]:
        """生成response对象                                                   \n
        :param url: 目标url
        :param mode: 'get', 'post' 中选择
        :param data: post方式要提交的数据
        :param kwargs: 连接参数
        :return: tuple，第一位为Response或None，第二位为出错信息或'Success'
        """
        url = quote(url, safe='/:&?=%;#@+!')
        kwargs = CaseInsensitiveDict(kwargs)
        session = copy_session(self.session)

        if 'headers' not in kwargs:
            kwargs['headers'] = {}
        else:
            kwargs['headers'] = CaseInsensitiveDict(kwargs['headers'])

        # 设置referer、host和timeout值
        parsed_url = urlparse(url)
        hostname = parsed_url.hostname
        scheme = parsed_url.scheme

        if not ('Referer' in kwargs['headers'] or 'Referer' in self.session.headers):
            kwargs['headers']['Referer'] = self._page.url if self._page is not None else f'{scheme}://{hostname}'
        if 'Host' not in kwargs['headers']:
            kwargs['headers']['Host'] = hostname
        if not ('timeout' in kwargs['headers'] or 'timeout' in self.session.headers):
            kwargs['timeout'] = self.timeout

        # 执行连接
        r = err = None
        for i in range(self.retry + 1):
            try:
                if mode == 'get':
                    r = session.get(url, **kwargs)
                elif mode == 'post':
                    r = session.post(url, data=data, **kwargs)

                if r:
                    return _set_charset(r), 'Success'

            except Exception as e:
                err = e

            if r and r.status_code in (403, 404):
                break
            if i < self.retry:
                sleep(self.interval)

        # 返回失败结果
        if r is None:
            return None, '连接失败' if err is None else err
        if not r.ok:
            return r, f'状态码：{r.status_code}'

    def _get_usable_thread(self) -> Union[int, None]:
        """获取可用线程，没有则返回None"""
        for k, v in self._threads.items():
            if v is None:
                return k

    def _stop_show(self) -> None:
        """设置停止打印的变量"""
        input()
        self._stop_printing = True

    def _when_mission_done(self, mission: Mission) -> None:
        """当任务完成时执行的操作                     \n
        :param mission: 完结的任务
        :return: None
        """
        if self.set_print().log_mode == 'all' or (self.set_print().log_mode == 'fail' and mission.result is False):
            print(f'{mission.data.url}\n{mission.result}\n{mission.info}\n')

        if self.set_log().log_mode == 'all' or (self.set_log().log_mode == 'fail' and mission.result is False):
            data = {'url': mission.data.url,
                    'path': mission.data.goal_path,
                    'rename': mission.data.rename,
                    'post_data': mission.data.post_data,
                    'kwargs': mission.data.kwargs}
            self._logger.add_data(data)

    def _download(self,
                  mission_or_task: Union[Mission, Task],
                  thread_id: int) -> None:
        """此方法是执行下载的线程方法，用于根据任务下载文件     \n
        :param mission_or_task: 下载任务对象
        :param thread_id: 线程号
        :return: None
        """
        if mission_or_task.is_done:
            return
        if mission_or_task.state == 'cancel':
            mission_or_task.state = 'done'
            return

        file_url = mission_or_task.data.url
        post_data = mission_or_task.data.post_data
        kwargs = mission_or_task.data.kwargs

        if isinstance(mission_or_task, Task):
            task = mission_or_task
            kwargs = CaseInsensitiveDict(kwargs)
            if 'headers' not in kwargs:
                kwargs['headers'] = {'Range': f"bytes={task.range[0]}-{task.range[1]}"}
            else:
                kwargs['headers']['Range'] = f"bytes={task.range[0]}-{task.range[1]}"

            mode = 'post' if post_data is not None or kwargs.get('json', None) else 'get'
            r, inf = self._connect(file_url, mode=mode, data=post_data, **kwargs)

            if r:
                _do_download(r, task, False)
            else:
                task.set_done(False, inf)

            return

        # ===================开始处理mission====================
        mission = mission_or_task
        mission.info = '下载中'
        mission.state = 'running'

        rename = mission.data.rename
        goal_path = mission.data.goal_path
        file_exists = mission.data.file_exists
        split = mission.data.split

        goal_Path = Path(goal_path)
        # 按windows规则去除路径中的非法字符
        goal_path = goal_Path.anchor + sub(r'[*:|<>?"]', '', goal_path.lstrip(goal_Path.anchor)).strip()
        goal_Path = Path(goal_path).absolute()
        goal_Path.mkdir(parents=True, exist_ok=True)
        goal_path = str(goal_Path)

        if file_exists == 'skip' and rename and (goal_Path / rename).exists():
            mission.file_name = rename
            mission.path = goal_Path / rename
            mission.set_done('skip', str(mission.path))
            return

        mode = 'post' if post_data is not None or kwargs.get('json', None) else 'get'
        r, inf = self._connect(file_url, mode=mode, data=post_data, **kwargs)

        if mission.is_done:
            return

        if not r:
            mission.break_mission(result=False, info=inf)
            return

        # -------------------获取文件信息-------------------
        file_info = _get_file_info(r, goal_path, rename, file_exists, self._lock)
        file_size = file_info['size']
        full_path = file_info['path']
        mission.path = full_path
        mission.file_name = full_path.name
        mission.size = file_size

        if file_info['skip']:
            mission.set_done('skip', str(mission.path))
            return

        # -------------------设置分块任务-------------------
        first = False
        if split and file_size and file_size > self.block_size and r.headers.get('Accept-Ranges') == 'bytes':
            first = True
            chunks = [[s, min(s + self.block_size, file_size)] for s in range(0, file_size, self.block_size)]
            chunks[-1][-1] = ''
            chunks_len = len(chunks)

            task1 = Task(mission, chunks[0], f'1/{chunks_len}')
            mission.tasks_count = chunks_len
            mission.tasks = []
            mission.tasks.append(task1)

            for ind, chunk in enumerate(chunks[1:], 2):
                task = Task(mission, chunk, f'{ind}/{chunks_len}')
                mission.tasks.append(task)
                self._run_or_wait(task)

        else:  # 不分块
            task1 = Task(mission, None, '1/1')
            mission.tasks.append(task1)

        self._threads[thread_id]['mission'] = task1
        _do_download(r, task1, first)


def _do_download(r: Response, task: Task, first: bool = False):
    """执行下载任务                                    \n
    :param r: Response对象
    :param task: 任务
    :param first: 是否第一个分块
    :return: None
    """
    if task.is_done or task.mission.is_done:
        return

    task.set_states(result=None, info='下载中', state='running')
    block_size = 65536  # 64k
    result = None

    try:
        if first:  # 分块时第一块
            r_content = r.iter_content(chunk_size=task.range[1])
            task.add_data(next(r_content), 0)

        else:
            if task.range is None:  # 不分块
                for chunk in r.iter_content(chunk_size=block_size):
                    if task.state in ('cancel', 'done'):
                        result = 'cancel'
                        break
                    if chunk:
                        task.add_data(chunk, None)

            elif task.range[1] == '':  # 结尾的数据块
                begin = task.range[0]
                for chunk in r.iter_content(chunk_size=block_size):
                    if task.state in ('cancel', 'done'):
                        result = 'cancel'
                        break
                    if chunk:
                        task.add_data(chunk, seek=begin)
                        begin += len(chunk)

            else:  # 有始末数字的数据块
                begin, end = task.range
                num = (end - begin) // block_size
                for ind, chunk in enumerate(r.iter_content(chunk_size=block_size), 1):
                    if task.state in ('cancel', 'done'):
                        result = 'cancel'
                        break
                    if chunk:
                        task.add_data(chunk, seek=begin)
                        if ind <= num:
                            begin += block_size

    except Exception as e:
        result, info = False, f'下载失败。{r.status_code} {e}'

    else:
        result = 'success' if result is None else result
        info = str(task.path)

    finally:
        r.close()

    task.set_done(result=result, info=info)



# -*- coding:utf-8 -*-
"""
@Author  :   g1879
@Contact :   g1879@qq.com
@File    :   _funcs.py
"""
from copy import copy
from os import path as os_PATH
from pathlib import Path
from random import randint
from re import search, sub
from threading import Lock
from time import time
from typing import Union
from urllib.parse import unquote

from requests import Session, Response


def copy_session(s: Session):
    """复制输入Session对象，返回一个新的"""
    new = Session()
    new.headers = s.headers.copy()
    new.cookies = s.cookies.copy()
    new.stream = True
    new.auth = s.auth
    new.proxies = dict(s.proxies).copy()
    new.params = copy(s.params)  #
    new.cert = s.cert
    new.max_redirects = s.max_redirects
    new.trust_env = s.trust_env
    new.verify = s.verify

    return new


class BlockSizeSetter(object):
    def __set__(self, block_size, val):
        if isinstance(val, int) and val > 0:
            size = val
        elif isinstance(val, str):
            units = {'b': 1, 'k': 1024, 'm': 1048576, 'g': 21474836480}
            num = int(val[:-1])
            unit = units.get(val[-1].lower(), None)
            if unit and num > 0:
                size = num * unit
            else:
                raise ValueError('单位只支持B、K、M、G，数字必须为大于0的整数。')
        else:
            raise TypeError('split_size只能传入int或str，数字必须为大于0的整数。')

        block_size._block_size = size

    def __get__(self, block_size, objtype=None) -> int:
        return block_size._block_size


class PathSetter(object):
    def __set__(self, goal_path, val):
        if val is not None and not isinstance(val, (str, Path)):
            raise TypeError('goal_path只能是str或Path类型。')
        goal_path._goal_path = str(val) if isinstance(val, Path) else val

    def __get__(self, goal_path, objtype=None):
        return goal_path._goal_path


class FileExistsSetter(object):
    def __set__(self, file_exists, mode):
        mode = mode.lower()
        if mode not in ('skip', 'overwrite', 'rename'):
            raise ValueError("file_exists参数只能传入'skip', 'overwrite', 'rename'")
        file_exists._file_exists = mode

    def __get__(self, file_exists, objtype=None):
        return file_exists._file_exists


class LogMode(object):
    """设置记录或打印哪些任务信息"""

    def __init__(self):
        self.log_mode = None

    def all(self):
        """全部"""
        self.log_mode = 'all'

    def none(self):
        """不记录或打印"""
        self.log_mode = None

    def fail(self):
        """只记录或打印失败任务"""
        self.log_mode = 'fail'


def get_usable_path(path: Union[str, Path]) -> Path:
    """检查文件或文件夹是否有重名，并返回可以使用的路径           \n
    :param path: 文件或文件夹路径
    :return: 可用的路径，Path对象
    """
    path = Path(path)
    parent = path.parent
    path = parent / make_valid_name(path.name)
    name = path.stem if path.is_file() else path.name
    ext = path.suffix if path.is_file() else ''

    first_time = True

    while path.exists():
        r = search(r'(.*)_(\d+)$', name)

        if not r or (r and first_time):
            src_name, num = name, '1'
        else:
            src_name, num = r.group(1), int(r.group(2)) + 1

        name = f'{src_name}_{num}'
        path = parent / f'{name}{ext}'
        first_time = None

    return path


def make_valid_name(full_name: str) -> str:
    """获取有效的文件名                  \n
    :param full_name: 文件名
    :return: 可用的文件名
    """
    # ----------------去除前后空格----------------
    full_name = full_name.strip()

    # ----------------使总长度不大于255个字符（一个汉字是2个字符）----------------
    r = search(r'(.*)(\.[^.]+$)', full_name)  # 拆分文件名和后缀名
    if r:
        name, ext = r.group(1), r.group(2)
        ext_long = len(ext)
    else:
        name, ext = full_name, ''
        ext_long = 0

    while get_long(name) > 255 - ext_long:
        name = name[:-1]

    full_name = f'{name}{ext}'

    # ----------------去除不允许存在的字符----------------
    return sub(r'[<>/\\|:*?\n]', '', full_name)


def get_long(txt) -> int:
    """返回字符串中字符个数（一个汉字是2个字符）          \n
    :param txt: 字符串
    :return: 字符个数
    """
    txt_len = len(txt)
    return int((len(txt.encode('utf-8')) - txt_len) / 2 + txt_len)


def _set_charset(response) -> Response:
    """设置Response对象的编码"""
    # 在headers中获取编码
    content_type = response.headers.get('content-type', '').lower()
    charset = search(r'charset[=: ]*(.*)?;', content_type)

    if charset:
        response.encoding = charset.group(1)

    # 在headers中获取不到编码，且如果是网页
    elif content_type.replace(' ', '').startswith('text/html'):
        re_result = search(b'<meta.*?charset=[ \\\'"]*([^"\\\' />]+).*?>', response.content)

        if re_result:
            charset = re_result.group(1).decode()
        else:
            charset = response.apparent_encoding

        response.encoding = charset

    return response


def _get_file_info(response,
                   goal_path: str = None,
                   rename: str = None,
                   file_exists: str = None,
                   lock: Lock = None) -> dict:
    """获取文件信息，大小单位为byte                   \n
    包括：size、path、skip
    :param response: Response对象
    :param goal_path: 目标文件夹
    :param rename: 重命名
    :param file_exists: 存在重名文件时的处理方式
    :param lock: 线程锁
    :return: 文件名、文件大小、保存路径、是否跳过
    """
    # ------------获取文件大小------------
    file_size = response.headers.get('Content-Length', None)
    file_size = None if file_size is None else int(file_size)

    # ------------获取网络文件名------------
    file_name = _get_file_name(response)

    # ------------获取保存路径------------
    goal_Path = Path(goal_path)
    # 按windows规则去除路径中的非法字符
    goal_path = goal_Path.anchor + sub(r'[*:|<>?"]', '', goal_path.lstrip(goal_Path.anchor)).strip()
    goal_Path = Path(goal_path).absolute()
    goal_Path.mkdir(parents=True, exist_ok=True)

    # ------------获取保存文件名------------
    # -------------------重命名，不改变扩展名-------------------
    if rename:
        ext_name = file_name.split('.')[-1]
        if '.' in rename or ext_name == file_name:  # 新文件名带后缀或原文件名没有后缀
            full_name = rename
        else:
            full_name = f'{rename}.{ext_name}'
    else:
        full_name = file_name

    full_name = make_valid_name(full_name)

    # -------------------生成路径-------------------
    skip = False
    full_path = goal_Path / full_name

    with lock:
        if full_path.exists():
            if file_exists == 'rename':
                full_path = get_usable_path(full_path)

            elif file_exists == 'skip':
                skip = True

            elif file_exists == 'overwrite':
                full_path.unlink()

        if not skip:
            with open(full_path, 'wb'):
                pass

    return {'size': file_size,
            'path': full_path,
            'skip': skip}


def _get_file_name(response) -> str:
    """从headers或url中获取文件名，如果获取不到，生成一个随机文件名
    :param response: 返回的response
    :return: 下载文件的文件名
    """
    file_name = ''
    charset = ''
    content_disposition = response.headers.get('content-disposition', '').replace(' ', '')

    # 使用header里的文件名
    if content_disposition:
        txt = search(r'filename\*="?([^";]+)', content_disposition)
        if txt:  # 文件名自带编码方式
            txt = txt.group(1).split("''", 1)
            if len(txt) == 2:
                charset, file_name = txt
            else:
                file_name = txt[0]

        else:  # 文件名没带编码方式
            txt = search(r'filename="?([^";]+)', content_disposition)
            if txt:
                file_name = txt.group(1)

                # 获取编码（如有）
                charset = response.encoding

        file_name = file_name.strip("'")

    # 在url里获取文件名
    if not file_name and os_PATH.basename(response.url):
        file_name = os_PATH.basename(response.url).split("?")[0]

    # 找不到则用时间和随机数生成文件名
    if not file_name:
        file_name = f'untitled_{time()}_{randint(0, 100)}'

    # 去除非法字符
    charset = charset or 'utf-8'
    return unquote(file_name, charset)



# -*- coding:utf-8 -*-
from pathlib import Path
from time import sleep, perf_counter
from typing import Union, List

from DataRecorder import ByteRecorder


class MissionData(object):
    """保存任务数据的对象"""

    def __init__(self, url: str,
                 goal_path: Union[str, Path],
                 rename: Union[str, None],
                 file_exists: str,
                 post_data: Union[str, dict, None],
                 split: bool,
                 kwargs: dict):
        self.url = url
        self.goal_path = goal_path
        self.rename = rename
        self.file_exists = file_exists
        self.post_data = post_data
        self.split = split
        self.kwargs = kwargs


class BaseTask(object):
    """任务类基类"""
    _DONE = 'done'

    def __init__(self, ID: Union[int, str]):
        """初始化                   \n
        :param ID: 任务id
        """
        self._id = ID
        self.state = 'waiting'  # 状态：'waiting'、'running'、'done'
        self.result = None  # 结果：'success'、'skip'、'cancel'、False、None
        self.info = '等待下载'  # 信息

    @property
    def id(self) -> Union[int, str]:
        """返回任务或子任务id"""
        return self._id

    @property
    def data(self):
        """返回任务数据"""
        return

    @property
    def is_done(self) -> bool:
        """返回任务是否结束"""
        return self.state == self._DONE

    def set_states(self,
                   result: Union[bool, None, str] = None,
                   info: str = None,
                   state: str = 'done') -> None:
        """设置任务结果值                                                  \n
        :param result: 结果：'success'、'skip'、'cancel'、False、None
        :param info: 任务信息
        :param state: 任务状态：'waiting'、'running'、'done'
        :return: None
        """
        self.result = result
        self.info = info
        self.state = state


class Mission(BaseTask):
    """任务类"""

    def __init__(self, ID: int, data: MissionData, download_kit):
        """初始化                               \n
        :param ID: 任务id
        :param data: 任务数据
        """
        super().__init__(ID)
        self.download_kit = download_kit
        self._data: MissionData = data
        self.size = None

        self.tasks: List[Task] = []  # 多线程下载单个文件时的子任务
        self.tasks_count = 1
        self.done_tasks_count = 0

        self.file_name = None
        self._path: Union[Path, None] = None  # 文件完整路径，Path对象
        self._recorder = None

    def __repr__(self) -> str:
        return f'<Mission {self.id} {self.info} {self.file_name}>'

    @property
    def data(self) -> MissionData:
        """返回任务数据"""
        return self._data

    @property
    def path(self) -> Union[str, Path]:
        """返回文件保存路径"""
        return self._path

    @path.setter
    def path(self, path: Union[str, Path, None]) -> None:
        """设置文件保存路径"""
        if isinstance(path, (Path, str)):
            path = Path(path)
            self.file_name = path.name

        self._path = path
        self.recorder.set_path(path)

    @property
    def recorder(self) -> ByteRecorder:
        """返回记录器对象"""
        if self._recorder is None:
            self._recorder = ByteRecorder(cache_size=100)
            self._recorder.show_msg = False
        return self._recorder

    @property
    def rate(self) -> Union[float, None]:
        """返回下载进度百分比"""
        if self.path and self.path.exists():
            return round((self.path.stat().st_size / self.size) * 100, 2) if self.size else None
        else:
            return

    def set_done(self, result: Union[bool, None, str], info: str) -> None:
        """设置一个任务为done状态                                          \n
        :param result: 结果：'success'、'skip'、'cancel'、False、None
        :param info: 任务信息
        :return: None
        """
        if result == 'skip':
            self.set_states(result=result, info=info, state=self._DONE)

        elif result == 'cancel' or result is False:
            self.recorder.clear()
            self.set_states(result=result, info=info, state=self._DONE)

        elif result == 'success':
            self.recorder.record()
            if self.size and self.path.stat().st_size < self.size:
                self.del_file()
                self.set_states(False, '下载失败', self._DONE)
            else:
                self.set_states('success', info, self._DONE)

        self.download_kit._when_mission_done(self)

    def a_task_done(self, is_success: bool, info: str) -> None:
        """当一个task完成时调用                 \n
        :param is_success: 该task是否成功
        :param info: 该task传入的信息
        :return: None
        """
        if self.is_done:
            return

        if is_success is False:
            self.break_mission(False, info)
            return

        self.done_tasks_count += 1
        if self.done_tasks_count == self.tasks_count:
            self.set_done('success', info)

    def break_mission(self, result: Union[bool, None, str], info: str) -> None:
        """中止该任务，停止未下载完的task                                  \n
        :param result: 结果：'success'、'skip'、'cancel'、False、None
        :param info: 任务信息
        :return: None
        """
        if self.is_done:
            return

        for task in self.tasks:
            if not task.is_done:
                task.set_states(result=result, info=info, state='cancel')

        while any((not i.is_done for i in self.tasks)):
            sleep(.1)

        self.set_done(result, info)
        self.del_file()

    def cancel(self) -> None:
        """取消该任务，停止未下载完的task"""
        self.break_mission('cancel', '已取消')

    def del_file(self):
        """删除下载的文件"""
        if self.path and self.path.exists():
            try:
                self.path.unlink()
            except Exception:
                pass

    def wait(self, show: bool = True, timeout: float = 0) -> tuple:
        """等待当前任务完成                  \n
        :param show: 是否显示下载进度
        :param timeout: 超时时间
        :return: 任务结果和信息组成的tuple
        """
        if show:
            print(f'url：{self.data.url}')
            t2 = perf_counter()
            while self.file_name is None and perf_counter() - t2 < 4:
                sleep(0.01)
            print(f'文件名：{self.file_name}')
            print(f'目标路径：{self.path}')
            if not self.size:
                print('未知大小 ', end='')

        t1 = perf_counter()
        while not self.is_done and (perf_counter() - t1 < timeout or timeout == 0):
            if show and self.size:
                try:
                    rate = round((self.path.stat().st_size / self.size) * 100, 2)
                    print(f'\r{rate}% ', end='')
                except FileNotFoundError:
                    pass

            sleep(0.1)

        if show:
            if self.result is False:
                print(f'下载失败 {self.info}')
            elif self.result == 'success':
                print('\r100% ', end='')
                print(f'下载完成 {self.info}')
            elif self.result == 'skip':
                print(f'已跳过 {self.info}')
            print()

        return self.result, self.info


class Task(BaseTask):
    """子任务类"""

    def __init__(self, mission: Mission, range_: Union[list, None], ID: str):
        """初始化                               \n
        :param mission: 父任务对象
        :param range_: 读取文件数据范围
        :param ID: 任务id
        """
        super().__init__(ID)
        self.mission = mission  # 父任务
        self.range = range_  # 分块范围

    def __repr__(self) -> str:
        return f'<Task M{self.mid} T{self._id}  {self.info} {self.file_name}>'

    @property
    def mid(self) -> int:
        """返回父任务id"""
        return self.mission.id

    @property
    def data(self) -> MissionData:
        """返回任务数据对象"""
        return self.mission.data

    @property
    def path(self) -> str:
        """返回文件保存路径"""
        return self.mission.path

    @property
    def file_name(self) -> str:
        """返回文件名"""
        return self.mission.file_name

    def add_data(self, data: bytes, seek: int = None) -> None:
        """把数据输入到记录器                           \n
        :param data: 文件字节数据
        :param seek: 在文件中的位置，None表示最后
        :return: None
        """
        self.mission.recorder.add_data(data, seek)

    def set_done(self, result: Union[bool, None, str], info: str) -> None:
        """设置一个子任务为done状态                                          \n
        :param result: 结果：'success'、'skip'、'cancel'、False、None
        :param info: 任务信息
        :return: None
        """
        self.set_states(result=result, info=info, state=self._DONE)
        self.mission.a_task_done(result, info)
